import azure.functions as func
import logging, os, tempfile, subprocess, json, glob, shutil, re
from pathlib import Path
from azure.storage.blob import BlobServiceClient

app = func.FunctionApp()

@app.blob_trigger(
    arg_name="myblob",
    path="input-container-name" + "/{name}." + "MF4",
    connection="StorageAccountConnectionStringInput",
)
def MdfToParquet(myblob: func.InputStream):
    logging.info(f"Trigger function processed blob " f"Name: {myblob.name} ")
    logging.info("Function version: v1.1.0")

    # define variables
    container_input = "input-container-name"
    container_output = "output-container-name"

    decoder = "mdf2parquet_decode"
    files_uploaded = 0
    dbc_groups = True

    # initialize
    blob_service_client_input = BlobServiceClient.from_connection_string(os.getenv("StorageAccountConnectionStringInput"))
    blob_service_client_output = BlobServiceClient.from_connection_string(os.getenv("StorageAccountConnectionStringOutput"))

    # Create a temporary directory to store the blob file
    with tempfile.TemporaryDirectory() as temp:
        tmp_dir = Path(temp)
        tmp_input_dir = tmp_dir / "input"
        tmp_input_dir.mkdir()
        tmp_output_dir = tmp_dir / "output"
        tmp_output_dir.mkdir()

        # extract device ID (note: Only works if the CANedge uploaded file structure is used)
        device_id = ""
        parts = myblob.name.split("/")

        # Check if the path has 4 parts and a valid device ID syntax in the 2nd position
        if len(parts) == 4 and re.match("[0-9A-F]{8}$", parts[1]):
            device_id = parts[1]
            logging.info(f"Device ID: {device_id}")
        else:
            logging.info(
                f"WARNING: Unable to extract device_id - log file does not use CANedge path"
            )

        # Download the dbc-groups.json file (if it exists) and use the device ID to extract the relevant DBC file list for the device
        # If no match is found, the script simply applies all DBC files across any device ID
        device_specific_dbc_files = []

        if dbc_groups and device_id != "":
            fs_dbc_groups_file_path = tmp_input_dir / "dbc-groups.json"
            try:
                dbc_blob_client = blob_service_client_input.get_blob_client(
                    container=container_input, blob="dbc-groups.json"
                )
                with open(fs_dbc_groups_file_path, "wb") as file:
                    file.write(dbc_blob_client.download_blob().readall())
                logging.info("Downloaded dbc-groups.json successfully.")

                # Read the JSON file
                with open(fs_dbc_groups_file_path, "r") as file:
                    data = json.load(file)

                # Iterate over the groups in the JSON
                for group in data["dbc_groups"]:
                    # Check if the device_id is in the current group
                    if device_id in group["devices"]:
                        # Add the dbc_files list to the variable
                        device_specific_dbc_files = group["dbc_files"]
                        logging.info(
                            "Device specific DBC files:", device_specific_dbc_files
                        )
                        break
            except Exception as e:
                logging.info(f"Note: Applying all DBC files across all devices")

        # Download DBC files
        dbc_files = []
        for type in ["can", "lin"]:
            container_client = blob_service_client_input.get_container_client(
                container=container_input
            )
            blob_list = container_client.list_blobs(name_starts_with=type)

            for blob in blob_list:
                dbc_object_name = blob.name
                if dbc_object_name.endswith(".dbc") and (
                    len(device_specific_dbc_files) == 0
                    or dbc_object_name in device_specific_dbc_files
                ):
                    dbc_files.append(dbc_object_name)
                    dbc_blob_client = blob_service_client_input.get_blob_client(
                        container=container_input, blob=dbc_object_name
                    )

                    dbc_file_path = tmp_input_dir / dbc_object_name
                    with open(dbc_file_path, "wb") as dbc_file:
                        dbc_file.write(dbc_blob_client.download_blob().readall())
                    logging.info(
                        f"Downloaded DBC file: {dbc_object_name} to {dbc_file_path}"
                    )

        if len(dbc_files) == 0:
            logging.info(
                f"ERROR: No DBC files found with valid prefix in {container_input}/ - exiting script"
            )
            return

        # ----------------------------

        # specify filename and path for MDF file and download it
        filename = os.path.basename(myblob.name)
        fs_logfile_path = tmp_input_dir / filename
        with open(fs_logfile_path, "wb") as file:
            data = myblob.read()
            file.write(data)

        logging.info(f"{filename} downloaded to {fs_logfile_path}")

        current_directory = os.getcwd()
        logging.info(
            f"Current working directory is {current_directory} and contains these files: {os.listdir(current_directory)}"
        )

        # copy decoder to tmp input dir
        decoder_path_root = Path(current_directory) / decoder
        decoder_target_path = tmp_input_dir / decoder
        shutil.copy(decoder_path_root, decoder_target_path)

        decoder_path = decoder_target_path

        all_files = glob.glob(f"{tmp_input_dir}/**/*", recursive=True)
        logging.info(f"Files/folders in tmp_input_dir: {all_files}")

        if decoder_path.exists():
            logging.info(f"Executable '{decoder}' is present in the directory.")
            try:
                result = subprocess.run(
                    [decoder_path, "-v"],
                    cwd=tmp_input_dir,
                    capture_output=True,
                    text=True,
                )
                logging.info(f"Decoder version: {result.stdout}")

                result = subprocess.run(
                    [
                        str(decoder_path),
                        "-i",
                        str(fs_logfile_path),
                        "-O",
                        str(tmp_output_dir),
                        "--verbosity=5",
                        "-X",
                    ],
                    cwd=str(tmp_input_dir),
                    check=True,
                    text=True,
                    capture_output=True,
                )
                logging.info(f"Subprocess output:")
                logging.info(f"{result.stdout}")
                logging.info(f"{result.stderr}")

                all_files = glob.glob(f"{tmp_output_dir}/**/*", recursive=True)
                logging.info(f"Files/folders in tmp_output_dir: {all_files}")

                # Initialize the ContainerClient for the output container
                container_client_output = blob_service_client_output.get_container_client(
                    container_output
                )

                # Loop through each Parquet file in the tmp_output_dir and upload to the output container
                for fs_output_file in tmp_output_dir.rglob("*.parquet"):
                    if fs_output_file.is_file():
                        container_output_path = str(
                            fs_output_file.relative_to(tmp_output_dir)
                        )

                        # Get a BlobClient to handle the upload
                        blob_client = container_client_output.get_blob_client(
                            blob=container_output_path
                        )

                        # Open the file and upload it
                        with open(fs_output_file, "rb") as data:
                            blob_client.upload_blob(data, overwrite=True)
                            logging.info(
                                f"Uploaded {fs_output_file.name} to {container_output} as {container_output_path}"
                            )

                        files_uploaded += 1

            except subprocess.CalledProcessError as e:
                logging.error(f"Error running the executable: {e}")
        else:
            logging.error(f"Executable '{decoder}' not found in the directory.")

    logging.info("Processing complete.")

    if files_uploaded > 0:
        logging.info(f"{files_uploaded} parquet files uploaded to {container_output}")
    else:
        logging.info(f"WARNING: No parquet files uploaded to {container_output}")
